<?php

declare(strict_types=1);

namespace Drupal\date_point\Plugin\Field\FieldType\DatePoint;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup as TM;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\date_point\ClockTrait;
use Drupal\date_point\Data\DatePointBehavior;
use Drupal\date_point\Data\Precision;

/**
 * Defines the 'date_point' entity field type.
 *
 * @FieldType(
 *   id = "date_point",
 *   label = @Translation("Date Point"),
 *   description = {
 *     @Translation("Ideal for using date and time calculations or comparisons."),
 *     @Translation("Date and time stored in RFC 3339 format using native DB data type."),
 *     @Translation("Compact and efficient for storage, sorting and calculations."),
 *   },
 *   category = "date_time",
 *   default_widget = "date_point_date_time_local",
 *   default_formatter = "date_point_default",
 *   list_class = "\Drupal\date_point\Plugin\Field\FieldType\DatePoint\DatePointItemList",
 * )
 *
 * @phpstan-type StanPrecision 'seconds'|'milliseconds'|'microseconds'
 * @phpstan-type StanDefaultValue 'now'|'none'
 * @phpstan-type SettingValue 'precision'|'default_value'|null
 *
 * @property string|null $value
 */
final class DatePointItem extends FieldItemBase {

  use ClockTrait;

  public const STORAGE_TIMEZONE = 'UTC';
  public const BASE_FORMAT = 'Y-m-d H:i:s';

  /**
   * @return array{precision: 'seconds'|'milliseconds'|'microseconds'}
   */
  public static function defaultStorageSettings(): array {
    return ['precision' => Precision::SECONDS->value];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{precision: array}
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data): array {
    $element['precision'] = [
      '#type' => 'radios',
      '#title' => new TM('Precision'),
      '#default_value' => $this->getSetting('precision'),
      '#options' => Precision::getOptions(),
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   *
   * @return non-empty-list<\Drupal\Core\StringTranslation\TranslatableMarkup>
   */
  public static function storageSettingsSummary(FieldStorageDefinitionInterface $storage_definition): array {
    return [
      new TM('Precision: @precision', ['@precision' => $storage_definition->getSetting('precision')]),
    ];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{behavior: 'manual'}
   */
  public static function defaultFieldSettings(): array {
    return ['behavior' => 'manual'];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{behavior: array}
   */
  public function fieldSettingsForm(array $form, FormStateInterface $form_state): array {
    $element['behavior'] = [
      '#type' => 'radios',
      '#title' => new TM('Behavior'),
      '#default_value' => $this->getSetting('behavior'),
      '#options' => DatePointBehavior::getOptions(),
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   *
   * @return non-empty-list<\Drupal\Core\StringTranslation\TranslatableMarkup>
   */
  public static function fieldSettingsSummary(FieldDefinitionInterface $field_definition): array {
    $behavior = DatePointBehavior::getOptions()[$field_definition->getSetting('behavior')];
    return [
      new TM('Behavior: @behavior', ['@behavior' => $behavior]),
    ];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{value: \Drupal\Core\TypedData\DataDefinition}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition): array {
    $properties['value'] = DataDefinition::create('date_point')
      ->setLabel(new TM('Date point value'))
      ->setRequired(TRUE);
    return $properties;
  }

  /**
   * {@inheritdoc}
   *
   * @return non-empty-list<\Symfony\Component\Validator\Constraint>
   */
  public function getConstraints(): array {
    $constraints = parent::getConstraints();
    $options = [
      'value' => [
        'NotBlank' => [],
        'DatePoint' => ['precision' => $this->getPrecision()],
      ],
    ];
    $constraints[] = \Drupal::typedDataManager()
      ->getValidationConstraintManager()
      ->create('ComplexData', $options);
    return $constraints;
  }

  /**
   * {@inheritdoc}
   *
   * @return array{columns: array{value: array}, indexes: array{value: array}}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition): array {
    $precision = Precision::from($field_definition->getSetting('precision'));
    $date_formatted = (new \DateTimeImmutable())->format(self::buildFormat($precision));
    $columns['value'] = [
      'type' => 'varchar',
      'description' => 'The date point value.',
      'mysql_type' => \sprintf('datetime(%d)', $precision->toInteger()),
      'pgsql_type' => \sprintf('timestamp(%d)', $precision->toInteger()),
      'length' => \strlen($date_formatted),
      'not null' => FALSE,
    ];
    return [
      'columns' => $columns,
      'indexes' => ['value' => ['value']],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty(): bool {
    $value = $this->get('value')->getValue();
    return $value === NULL || $value === '';
  }

  /**
   * {@inheritdoc}
   *
   * The implementation is a bit tricky because timestamp with microseconds may
   * exceed PHP_FLOAT_MAX. So we first create a date with only microseconds then
   * modify it with integer timestamp.
   *
   * @return array{value: non-empty-string}
   */
  public static function generateSampleValue(FieldDefinitionInterface $field_definition): array {
    $microseconds = \mt_rand(0, 1_000_000) / 1_000_000;
    $date = \DateTimeImmutable::createFromFormat('U.u', (string) \round($microseconds, 6));

    $year = 60 * 60 * 24 * 365;
    $timestamp = (new \DateTimeImmutable())->getTimestamp() + \mt_rand(-$year, $year);
    $modifier = \sprintf('%d seconds', $timestamp);

    $format = self::buildFormat(
      Precision::from($field_definition->getSetting('precision')),
    );
    return [
      'value' => $date->modify($modifier)->format($format),
    ];
  }

  /**
   * {@selfdoc}
   */
  public function getPrecision(): Precision {
    return Precision::from($this->getSetting('precision'));
  }

  /**
   * {@selfdoc}
   */
  public function getFormat(): string {
    return self::buildFormat($this->getPrecision());
  }

  /**
   * {@selfdoc}
   */
  public static function buildFormat(Precision $precision): string {
    return match ($precision) {
      Precision::SECONDS => self::BASE_FORMAT,
      Precision::MILLISECONDS => self::BASE_FORMAT . '.v',
      Precision::MICROSECONDS => self::BASE_FORMAT . '.u',
    };
  }

  /**
   * {@selfdoc}
   */
  public function getDatePoint(): \DateTimeImmutable {
    return new \DateTimeImmutable($this->value, new \DateTimeZone(self::STORAGE_TIMEZONE));
  }

  /**
   * {@selfdoc}
   */
  public function setDatePoint(?\DateTimeImmutable $datetime = NULL): void {
    $datetime ??= self::getClock()
      ->now()
      ->setTimezone(new \DateTimeZone(self::STORAGE_TIMEZONE));
    $this->setValue(['value' => $datetime->format($this->getFormat())]);
  }

}
