<?php

declare(strict_types=1);

namespace Drupal\date_point\Plugin\Field\FieldType\Time;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup as TM;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\date_point\ClockTrait;
use Drupal\date_point\Data\Precision;
use Drupal\date_point\Data\Time;

/**
 * Defines the 'time' entity field type.
 *
 * @FieldType(
 *   id = "time",
 *   label = @Translation("Time"),
 *   description = {
 *     @Translation("Ideal for using time calculations or comparisons."),
 *     @Translation("Date and time stored using native DB time type."),
 *     @Translation("Compact and efficient for storage, sorting and calculations."),
 *   },
 *   category = "date_time",
 *   default_widget = "time",
 *   default_formatter = "time",
 *   list_class = "\Drupal\date_point\Plugin\Field\FieldType\Time\TimeItemList",
 * )
 *
 * @phpstan-type StanPrecision 'seconds'|'milliseconds'|'microseconds'
 *
 * @property string|null $value
 */
final class TimeItem extends FieldItemBase {

  use ClockTrait;

  public const STORAGE_TIMEZONE = 'UTC';
  public const BASE_FORMAT = 'H:i:s';

  /**
   * @return array{precision: 'seconds'|'milliseconds'|'microseconds'}
   */
  public static function defaultStorageSettings(): array {
    return ['precision' => Precision::SECONDS->value];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{precision: array}
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data): array {
    $element['precision'] = [
      '#type' => 'radios',
      '#title' => new TM('Precision'),
      '#default_value' => $this->getSetting('precision'),
      '#options' => Precision::getOptions(),
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   *
   * @return non-empty-list<\Drupal\Core\StringTranslation\TranslatableMarkup>
   */
  public static function storageSettingsSummary(FieldStorageDefinitionInterface $storage_definition): array {
    return [
      new TM('Precision: @precision', ['@precision' => $storage_definition->getSetting('precision')]),
    ];
  }

  /**
   * {@inheritdoc}
   *
   * @return array{value: \Drupal\Core\TypedData\DataDefinition}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition): array {
    $properties['value'] = DataDefinition::create('time')
      ->setLabel(new TM('Time value'))
      ->setRequired(TRUE);
    return $properties;
  }

  /**
   * {@inheritdoc}
   *
   * @return non-empty-list<\Symfony\Component\Validator\Constraint>
   */
  public function getConstraints(): array {
    $constraints = parent::getConstraints();
    $options = [
      'value' => [
        'NotBlank' => [],
        'Time' => [],
      ],
    ];
    $constraints[] = \Drupal::typedDataManager()
      ->getValidationConstraintManager()
      ->create('ComplexData', $options);
    return $constraints;
  }

  /**
   * {@inheritdoc}
   *
   * @return array{columns: array{value: array}, indexes: array{value: array}}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition): array {
    $precision = Precision::from($field_definition->getSetting('precision'));
    $native_type = \sprintf('time(%d)', $precision->toInteger());
    $date_formatted = (new Time('00:00'))->format(self::buildFormat($precision));
    $columns['value'] = [
      'type' => 'varchar',
      'description' => 'The time value.',
      'mysql_type' => $native_type,
      'pgsql_type' => $native_type,
      'length' => \strlen($date_formatted),
      'not null' => FALSE,
    ];
    return [
      'columns' => $columns,
      'indexes' => ['value' => ['value']],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty(): bool {
    $value = $this->get('value')->getValue();
    return $value === NULL || $value === '';
  }

  /**
   * {@inheritdoc}
   *
   * @return array{value: non-empty-string}
   */
  public static function generateSampleValue(FieldDefinitionInterface $field_definition): array {
    $time = Time::fromSeconds(
      \mt_rand(1, 24 * 3_600) - 1 + \mt_rand() / \mt_getrandmax(),
    );
    $format = self::buildFormat(
      Precision::from($field_definition->getSetting('precision')),
    );
    return [
      'value' => $time->format($format),
    ];
  }

  /**
   * {@selfdoc}
   */
  public function getPrecision(): Precision {
    return Precision::from($this->getSetting('precision'));
  }

  /**
   * {@selfdoc}
   */
  public function getFormat(): string {
    return self::buildFormat($this->getPrecision());
  }

  /**
   * {@selfdoc}
   */
  public static function buildFormat(Precision $precision): string {
    return match ($precision) {
      Precision::SECONDS => self::BASE_FORMAT,
      Precision::MILLISECONDS => self::BASE_FORMAT . '.v',
      Precision::MICROSECONDS => self::BASE_FORMAT . '.u',
    };
  }

  /**
   * {@selfdoc}
   */
  public function getTime(): Time {
    return new Time($this->value);
  }

  /**
   * {@selfdoc}
   */
  public function setTime(?Time $time = NULL): void {
    if ($time === NULL) {
      $date_time = self::getClock()
        ->now()
        ->setTimezone(new \DateTimeZone(self::STORAGE_TIMEZONE));
      $time = Time::fromDateTime($date_time);
    }
    $this->setValue(['value' => $time->format($this->getFormat())]);
  }

}
