<?php

declare(strict_types=1);

namespace Drupal\date_point\Plugin\Validation\Constraint;

use Drupal\date_point\Plugin\Field\FieldType\DatePoint\DatePointItem;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;
use Symfony\Component\Validator\Exception\UnexpectedTypeException;
use Symfony\Component\Validator\Exception\UnexpectedValueException;

/**
 * Validates the date point constraint.
 */
final class DatePointConstraintValidator extends ConstraintValidator {

  /**
   * {@inheritdoc}
   */
  public function validate(mixed $value, Constraint $constraint): void {
    if (!$constraint instanceof DatePointConstraint) {
      throw new UnexpectedTypeException($constraint, DatePointConstraint::class);
    }

    if (!\is_scalar($value) && !$value instanceof \Stringable) {
      throw new UnexpectedValueException($value, 'string');
    }

    $value = (string) $value;

    $this->validateFormat($value, $constraint);
    // There is no point to validate range if date is not correct.
    if (\count($this->context->getViolations()) === 0) {
      $this->validateRange($value, $constraint);
    }
  }

  /**
   * {@selfdoc}
   */
  private function validateFormat(string $value, DatePointConstraint $constraint): void {
    // @phpstan-ignore-next-line
    \DateTimeImmutable::createFromFormat(
      DatePointItem::buildFormat($constraint->precision), $value,
    );

    $result = \DateTimeImmutable::getLastErrors();
    if (!$result) {
      return;
    }

    if ($constraint->verbose) {
      /** @var non-empty-list<string> $all_errors */
      $all_errors = [...$result['errors'], ... $result['warnings']];
      $error = \mb_strtolower(\rtrim(\reset($all_errors), '.'));
      // Display just one error.
      $this->context->addViolation(
        $constraint->verboseMessage, ['%error' => $error],
      );
    }
    else {
      $this->context->addViolation($constraint->message);
    }
  }

  /**
   * {@selfdoc}
   */
  private function validateRange(string $value, DatePointConstraint $constraint): void {
    if ($constraint->min && $value < $constraint->min) {
      $this->context->addViolation($constraint->minMessage, ['@min' => $constraint->min]);
    }
    if ($constraint->max && $value > $constraint->max) {
      $this->context->addViolation($constraint->maxMessage, ['@max' => $constraint->max]);
    }
  }

}
