<?php

declare(strict_types=1);

namespace Drupal\Tests\date_point\Kernel\Plugin\FieldType;

use Drupal\date_point\Data\DatePointBehavior as Behavior;
use Drupal\date_point\Data\Precision;
use Drupal\entity_test\Entity\EntityTest;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\Tests\field\Kernel\FieldKernelTestBase;

/**
 * {@selfdoc}
 */
final class DatePointItemListTest extends FieldKernelTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['date_point', 'date_point_test'];

  /**
   * {@selfdoc}
   */
  public function testPreSave(): void {
    $clock = $this->container->get('date_point_test.state_clock');

    // -- 'manual' behavior.
    self::createField('field_example_m', 'milliseconds', 'manual');
    $entity = EntityTest::create();
    $entity->save();
    self::assertSame([], $entity->get('field_example_m')->getValue());
    $entity->save();
    self::assertSame([], $entity->get('field_example_m')->getValue());

    // -- 'created' behavior.
    self::createField('field_example_c', 'milliseconds', 'created');
    $entity = EntityTest::create();
    $clock->set(new \DateTimeImmutable('2025-10-04 00:00:00.123', new \DateTimeZone('UTC')));
    $entity->save();
    self::assertSame([['value' => '2025-10-04 00:00:00.123']], $entity->get('field_example_c')->getValue());
    $clock->set(new \DateTimeImmutable('2025-10-04 00:10:00.123', new \DateTimeZone('UTC')));
    $entity->save();
    // The value should not be changed as the entity was not new.
    self::assertSame([['value' => '2025-10-04 00:00:00.123']], $entity->get('field_example_c')->getValue());

    // -- 'updated' behavior.
    self::createField('field_example_u', 'milliseconds', 'updated');
    $entity = EntityTest::create();
    $clock->set(new \DateTimeImmutable('2025-10-04 00:00:00.123', new \DateTimeZone('UTC')));
    $entity->save();
    self::assertSame([['value' => '2025-10-04 00:00:00.123']], $entity->get('field_example_u')->getValue());
    $clock->set(new \DateTimeImmutable('2025-10-04 00:10:00.123', new \DateTimeZone('UTC')));
    $entity->save();
    // The value should be changed because of 'updated' behavior.
    self::assertSame([['value' => '2025-10-04 00:10:00.123']], $entity->get('field_example_u')->getValue());

    $clock->reset();
  }

  /**
   * {@selfdoc}
   */
  public function testGetPrecision(): void {
    self::createField('field_example_0', 'seconds', 'manual');
    $item_list = EntityTest::create()->get('field_example_0');
    self::assertSame(Precision::SECONDS, $item_list->getPrecision());

    self::createField('field_example_3', 'milliseconds', 'manual');
    $item_list = EntityTest::create()->get('field_example_3');
    self::assertSame(Precision::MILLISECONDS, $item_list->getPrecision());

    self::createField('field_example_6', 'microseconds', 'manual');
    $item_list = EntityTest::create()->get('field_example_6');
    self::assertSame(Precision::MICROSECONDS, $item_list->getPrecision());
  }

  /**
   * {@selfdoc}
   */
  public function testGetBehavior(): void {
    self::createField('field_example_m', 'seconds', 'manual');
    $item_list = EntityTest::create()->get('field_example_m');
    self::assertSame(Behavior::MANUAL, $item_list->getBehavior());

    self::createField('field_example_c', 'seconds', 'created');
    $item_list = EntityTest::create()->get('field_example_c');
    self::assertSame(Behavior::CREATED, $item_list->getBehavior());

    self::createField('field_example_u', 'seconds', 'updated');
    $item_list = EntityTest::create()->get('field_example_u');
    self::assertSame(Behavior::UPDATED, $item_list->getBehavior());
  }

  /**
   * {@selfdoc}
   */
  public function testGetDatePoint(): void {
    self::createField('field_example', 'seconds', 'manual');
    $entity = EntityTest::create();
    self::assertNull($entity->get('field_example')->getDatePoint());

    $entity->set('field_example', '2023-03-12 22:15:28');
    $date_point = $entity->get('field_example')->getDatePoint();
    self::assertSame('2023-03-12T22:15:28+00:00', $date_point->format(\DateTimeInterface::ATOM));
  }

  /**
   * {@selfdoc}
   */
  public function testSetDatePoint(): void {
    self::createField('field_example', 'milliseconds', 'manual');
    $item_list = EntityTest::create()->get('field_example');

    $item_list->setDatePoint(new \DateTimeImmutable('2021-10-04 23:10:00.123'));
    self::assertSame('2021-10-04 23:10:00.123', $item_list->value);

    $clock = $this->container->get('date_point_test.state_clock');
    $clock->set(new \DateTimeImmutable('2025-10-04 14:10:00.555', new \DateTimeZone('UTC')));
    $item_list->setDatePoint();
    self::assertSame('2025-10-04 14:10:00.555', $item_list->value);
    $clock->reset();
  }

  /**
   * {@selfdoc}
   */
  private static function createField(string $field_name, string $precision, string $behavior): void {
    $storage_values = [
      'entity_type' => 'entity_test',
      'field_name' => $field_name,
      'type' => 'date_point',
      'settings' => ['precision' => $precision],
    ];
    FieldStorageConfig::create($storage_values)->save();
    $field_values = [
      'entity_type' => 'entity_test',
      'field_name' => $field_name,
      'bundle' => 'entity_test',
      'settings' => ['behavior' => $behavior],
    ];
    FieldConfig::create($field_values)->save();
  }

}
